# Copyright (C) 2001-2003 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# $Id: message.py,v 1.2 2003/05/29 00:31:40 halley Exp $

"""DNS Messages"""

import cStringIO
import random
import string
import struct
import time

import DNS.flags
import DNS.name
import DNS.node
import DNS.opcode
import DNS.rcode
import DNS.rdata
import DNS.rdataclass
import DNS.rdataset
import DNS.rdatatype
import DNS.tsig

class ShortHeader(Exception):
    """Raised if the DNS packet passed to message.from_wire() is too short."""
    pass

class TrailingJunk(Exception):
    """Raised if the DNS packet passed to message.from_wire() has extra junk
    at the end of it."""
    pass

class UnknownHeaderField(Exception):
    """Raised if a header field name is not recognized when converting from
    text into a message."""
    pass

class UnknownHeaderFlag(Exception):
    """Raised if the text from of a flag is not recognized."""
    pass

class BadEDNS(Exception):
    """Raised if an OPT record occurs somewhere other than the start of
    the additional data section."""
    pass

class BadTSIG(Exception):
    """Raised if a TSIG record occurs somewhere other than the end of
    the additional data section."""
    pass

class Message(object):
    """A DNS message."""
    def __init__(self):
        self.id = random.randint(0, 65535)
        self.flags = 0
        self.question = []
        self.answer = []
        self.authority = []
        self.additional = []
        self.wire = ''
        self.current = 0
        self.last_name = None
        self.ednsflags = 0
        self.payload = 0
        self.edns = -1
        self.keyring = None
        self.keyname = None
        self.request_mac = ''
        self.other_data = ''
        self.tsig_error = 0
        self.fudge = 300
        self.mac = ''
        self.zone_rdclass = DNS.rdataclass.IN
        self.xfr = False
        self.origin = None
        self.tsig_ctx = None
        self.multi = False
        self.had_tsig = False

    #
    # String methods
    #
    
    def __repr__(self):
        return '<DNS message, ID ' + `self.id` + '>'
    
    def __str__(self):
        s = cStringIO.StringIO()
        print >> s, 'id %d' % self.id
        print >> s, 'opcode %s' % \
              DNS.opcode.to_text(DNS.opcode.from_flags(self.flags))
        rc = DNS.rcode.from_flags(self.flags, self.ednsflags)
        print >> s, 'rcode %s' % DNS.rcode.to_text(rc)
        print >> s, 'flags %s' % DNS.flags.to_text(self.flags)
        if self.edns >= 0:
            print >> s, 'edns %s' % self.edns
            if self.ednsflags != 0:
                print >> s, 'eflags %s' % \
                      DNS.flags.edns_to_text(self.ednsflags)
            print >> s, 'payload', self.payload
        is_update = DNS.opcode.is_update(self.flags)
        if is_update:
            print >> s, ';ZONE'
        else:
            print >> s, ';QUESTION'
        for n in self.question:
            print >> s, n
        if is_update:
            print >> s, ';PREREQ'
        else:
            print >> s, ';ANSWER'
        for n in self.answer:
            print >> s, n
        if is_update:
            print >> s, ';UPDATE'
        else:
            print >> s, ';AUTHORITY'
        for n in self.authority:
            print >> s, n
        print >> s, ';ADDITIONAL'
        for n in self.additional:
            print >> s, n
        #
        # We strip off the final \n so the caller can print the result without
        # doing weird things to get around eccentricities in Python print
        # formatting
        #
        return s.getvalue()[:-1]

    #
    # Comparison methods
    #

    def __eq__(self, other):
        """Two messages are equal if they have the same content in the
        header, question, answer, and authority sections."""
        if self.id != other.id:
            return False
        if self.flags != other.flags:
            return False
        for n in self.question:
            if n not in other.question:
                return False
        for n in other.question:
            if n not in self.question:
                return False
        for n in self.answer:
            if n not in other.answer:
                return False
        for n in other.answer:
            if n not in self.answer:
                return False
        for n in self.authority:
            if n not in other.authority:
                return False
        for n in other.authority:
            if n not in self.authority:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_response(self, other):
        """Is other a response to self?"""
        if other.flags & DNS.flags.QR == 0 or \
           self.id != other.id or \
           DNS.opcode.from_flags(self.flags) != \
           DNS.opcode.from_flags(other.flags):
            return False
        if DNS.rcode.from_flags(other.flags, other.ednsflags) != \
               DNS.rcode.NOERROR:
            return True
        if DNS.opcode.is_update(self.flags):
            return True
        for n in self.question:
            if n not in other.question:
                return False
        for n in other.question:
            if n not in self.question:
                return False
        return True
        
    def find_node(self, section, name, create = False, force_unique = False):
        """Find the node named name in the specified section.  If
        create is non-zero, create it if not found, unless force_unique is
        also non-zero, in which case always create a new node."""
        if create and force_unique:
            n = DNS.node.Node(name)
            section.append(n)
            return n
        for n in section:
            if n.name == name:
                return n
        if not create:
            raise KeyError
        n = DNS.node.Node(name)
        section.append(n)
        return n
        
    def get_question(self, qcount):
        """Read the next qcount records from the wire data and add them to
        the question section."""
        for i in xrange(0, qcount):
            (qname, used) = DNS.name.from_wire(self.wire, self.current)
            if not self.origin is None:
                qname = qname.relativize(self.origin)
            self.current = self.current + used
            (rdtype, rdclass) = \
                     struct.unpack('!HH',
                                   self.wire[self.current:self.current + 4])
            self.current = self.current + 4
            n = self.find_node(self.question, qname, True, True)
            rds = n.find_rdataset(rdclass, rdtype, DNS.rdatatype.NONE, True,
                                  True)
        
    def get_section(self, section, count):
        """Read the next count records from the wire data and add them to
        the specified section."""
        if DNS.opcode.is_update(self.flags):
            updating = True
            force_unique = True
        else:
            updating = False
            force_unique = False
        seen_opt = False
        for i in xrange(0, count):
            rr_start = self.current
            (nm, used) = DNS.name.from_wire(self.wire, self.current)
            if not self.origin is None:
                nm = nm.relativize(self.origin)
            self.current = self.current + used
            (rdtype, rdclass, ttl, rdlen) = \
                     struct.unpack('!HHIH',
                                   self.wire[self.current:self.current + 10])
            self.current = self.current + 10
            if rdtype == DNS.rdatatype.OPT:
                if not section is self.additional or seen_opt:
                    raise BadEDNS
                self.payload = rdclass
                self.ednsflags = ttl
                self.edns = (ttl & 0xff0000) >> 16
                seen_opt = True
            elif rdtype == DNS.rdatatype.TSIG:
                if not (section is self.additional and i == (count - 1)):
                    raise BadTSIG
                if not self.keyring is None:
                    self.tsig_ctx = DNS.tsig.validate(self.wire,
                                                      nm,
                                                      self.keyring[nm],
                                                      int(time.time()),
                                                      self.request_mac,
                                                      rr_start,
                                                      self.current,
                                                      rdlen,
                                                      self.tsig_ctx,
                                                      self.multi)
                self.had_tsig = True
            else:
                if ttl < 0:
                    ttl = 0
                if updating and \
                   (rdclass == DNS.rdataclass.ANY or
                    rdclass == DNS.rdataclass.NONE):
                    deleting = rdclass
                    rdclass = self.zone_rdclass
                else:
                    deleting = False
                if deleting == DNS.rdataclass.ANY:
                    covers = DNS.rdatatype.NONE
                    rd = None
                else:
                    rd = DNS.rdata.from_wire(rdclass, rdtype, self.wire,
                                             self.current, rdlen, self.origin)
                    covers = rd.covers()
                if self.xfr and rdtype == DNS.rdatatype.SOA:
                    force_unique = True
                n = self.find_node(section, nm, True, force_unique)
                rds = n.find_rdataset(rdclass, rdtype, covers, True,
                                      force_unique, deleting)
                if not rd is None:
                    rds.update_ttl(ttl)
                    rds.add(rd)
            self.current = self.current + rdlen

    def to_wire(self):
        """Return a string containing the message in DNS compressed wire
        format."""
        f = cStringIO.StringIO()
        compress = {}
        header = '\x00' * 12
        f.write(header)
        qcount = 0
        ancount = 0
        aucount = 0
        adcount = 0
        for n in self.question:
            qcount += n.to_wire(f, compress, None, True)
        for n in self.answer:
            ancount += n.to_wire(f, compress, None, False)
        for n in self.authority:
            aucount += n.to_wire(f, compress, None, False)
        if self.edns >= 0:
            stuff = struct.pack('!BHHIH', 0, DNS.rdatatype.OPT, self.playload,
                                self.ednsflags, 0)
            file.write(stuff)
            adcount += 1
        for n in self.additional:
            adcount += n.to_wire(f, compress, None, False)
        header = struct.pack('!HHHHHH', self.id, self.flags, qcount, ancount,
                             aucount, adcount)
        f.seek(0)
        f.write(header)
        if not self.keyname is None:
            secret = self.keyring[self.keyname]
            s = f.getvalue()
            (tsig_rdata, self.mac, ctx) = DNS.tsig.hmac_md5(s, self.keyname,
                                                            secret,
                                                            int(time.time()),
                                                            self.fudge,
                                                            self.id,
                                                            self.tsig_error,
                                                            self.other_data,
                                                            self.request_mac)
            f.seek(0, 2)
            self.keyname.to_wire(f, compress)
            stuff = struct.pack('!HHIH', DNS.rdatatype.TSIG,
                                DNS.rdataclass.ANY, 0, 0)
            f.write(stuff)
            start = f.tell()
            f.write(tsig_rdata)
            end = f.tell()
            assert end - start < 65536
            f.seek(start - 2)
            f.write(struct.pack('!H', end - start))
            adcount += 1
            stuff = struct.pack('!H', adcount)
            f.seek(10)
            f.write(stuff)
        s = f.getvalue()
        return s

    def use_tsig(self, keyring, keyname=None):
        """When sending, a TSIG signature using the specified keyring and
        keyname should be added.  If the keyname is left unspecified, the
        first key in the keyring will be used."""
        self.keyring = keyring
        if keyname is None:
            self.keyname = self.keyring.keys()[0]
        else:
            self.keyname = keyname

    def use_edns(self, edns, ednsflags, payload):
        """Configure EDNS behavior.  The edns parameter is the EDNS level
        to use.  Specifying None or -1 means 'do not use EDNS'."""
        if edns is None:
            edns = -1
        self.edns = edns
        self.ednsflags = ednsflags
        self.payload = payload

    def rcode(self):
        return DNS.rcode.from_flags(self.flags, self.ednsflags)

def from_wire(wire, keyring=None, request_mac='', xfr=False, origin=None,
              tsig_ctx = None, multi = False):
    """Convert a possibly compressed DNS wire format message into a message
    object.  If not None, keyring will be used to look up any TSIG keys.
    If not the empty string, request_mac will be used if the wire message
    is a response and is signed with a TSIG key."""
    m = Message()
    m.keyring = keyring
    m.request_mac = request_mac
    m.xfr = xfr
    m.origin = origin
    m.tsig_ctx = tsig_ctx
    m.multi = multi
    m.wire = wire
    l = len(m.wire)
    if l < 12:
        raise ShortHeader
    (m.id, m.flags, qcount, ancount, aucount, adcount) = \
         struct.unpack('!HHHHHH', m.wire[:12])
    m.current = 12
    m.get_question(qcount)
    m.get_section(m.answer, ancount)
    m.get_section(m.authority, aucount)
    m.get_section(m.additional, adcount)
    if m.current != l:
        raise TrailingJunk
    if m.multi and m.tsig_ctx:
        if not m.had_tsig:
            m.tsig_ctx.update(m.wire)
    return m

def header_line(m, tok, section):
    """Process one line from the text format header section."""
    (ttype, what) = tok.get()
    if what == 'id':
        m.id = tok.get_int()
    elif what == 'flags':
        while True:
            token = tok.get()
            if token[0] != DNS.tokenizer.IDENTIFIER:
                tok.unget(token)
                break
            m.flags = m.flags | DNS.flags.from_text(string.join(token[1]))
    elif what == 'edns':
        m.edns = tok.get_int()
        m.ednsflags = m.ednsflags | (m.edns << 16)
    elif what == 'eflags':
        if m.edns < 0:
            m.edns = 0
        while True:
            token = tok.get()
            if token[0] != DNS.tokenizer.IDENTIFIER:
                tok.unget(token)
                break
            m.ednsflags = m.ednsflags | \
                          DNS.flags.edns_from_text(string.join(token[1]))
    elif what == 'payload':
        m.payload = tok.get_int()
        if m.edns < 0:
            m.edns = 0
    elif what == 'opcode':
        text = tok.get_string()
        m.flags = m.flags | \
                  DNS.opcode.to_flags(DNS.opcode.from_text(text))
    elif what == 'rcode':
        text = tok.get_string()
        (value, evalue) = DNS.rcode.to_flags(rcode.from_text(text))
        m.flags = m.flags | value
        m.ednsflags = m.ednsflags | evalue
        if m.ednsflags != 0 and m.edns < 0:
            m.edns = 0
    else:
        raise UnknownHeaderField
    tok.get_eol()

def question_line(m, tok, section):
    """Process one line from the text format question section."""
    token = tok.get(want_leading = True)
    if token[0] != DNS.tokenizer.WHITESPACE:
        m.last_name = DNS.name.from_text(token[1])
    name = m.last_name
    token = tok.get()
    if token[0] != DNS.tokenizer.IDENTIFIER:
        raise SyntaxError
    # Class
    try:
        rdclass = DNS.rdataclass.from_text(token[1])
        token = tok.get()
        if token[0] != DNS.tokenizer.IDENTIFIER:
            raise SyntaxError
    except SyntaxError:
        raise SyntaxError
    except:
        rdclass = DNS.rdataclass.IN
    # Type
    rdtype = DNS.rdatatype.from_text(token[1])
    n = m.find_node(section, name, True, True)
    rds = n.find_rdataset(rdclass, rdtype, DNS.rdatatype.NONE, True)
    if DNS.opcode.is_update(m.flags):
        m.zone_rdclass = rdclass
    tok.get_eol()
        
def rr_line(m, tok, section):
    """Process one line from the text format answer, authority, or additional
    data sections."""
    if DNS.opcode.is_update(m.flags):
        updating = True
    else:
        updating = False
    deleting = False
    # Name
    token = tok.get(want_leading = True)
    if token[0] != DNS.tokenizer.WHITESPACE:
        m.last_name = DNS.name.from_text(token[1])
    name = m.last_name
    token = tok.get()
    if token[0] != DNS.tokenizer.IDENTIFIER:
        raise SyntaxError
    # TTL
    try:
        ttl = int(token[1], 0)
        token = tok.get()
        if token[0] != DNS.tokenizer.IDENTIFIER:
            raise SyntaxError
    except SyntaxError:
        raise SyntaxError
    except:
        ttl = 0
    # Class
    try:
        rdclass = DNS.rdataclass.from_text(token[1])
        token = tok.get()
        if token[0] != DNS.tokenizer.IDENTIFIER:
            raise SyntaxError
        if rdclass == DNS.rdataclass.ANY or rdclass == DNS.rdataclass.NONE:
            deleting = rdclass
            rdclass = m.zone_rdclass
    except SyntaxError:
        raise SyntaxError
    except:
        rdclass = DNS.rdataclass.IN
    # Type
    rdtype = DNS.rdatatype.from_text(token[1])
    n = m.find_node(section, name, True, updating)
    token = tok.get()
    if token[0] != DNS.tokenizer.EOL and token[0] != DNS.tokenizer.EOF:
        tok.unget(token)
        rd = DNS.rdata.from_text(rdclass, rdtype, tok, DNS.name.root)
        covers = rd.covers()
    else:
        rd = None
        covers = DNS.rdatatype.NONE
    rds = n.find_rdataset(rdclass, rdtype, covers, True, updating, deleting)
    if not rd is None:
        rds.update_ttl(ttl)
        rds.add(rd)
    
def from_file(f):
    """Read the next text format message from the specified file."""
    m = Message()
    tok = DNS.tokenizer.Tokenizer(f)
    line_method = header_line
    section = None
    while 1:
        token = tok.get(True, True)
        if token[0] == DNS.tokenizer.EOL or token[0] == DNS.tokenizer.EOF:
            break
        if token[0] == DNS.tokenizer.COMMENT:
            u = token[1].upper()
            if u == 'HEADER':
                line_method = header_line
            elif u == 'QUESTION' or u == 'ZONE':
                line_method = question_line
                section = m.question
            elif u == 'ANSWER' or u == 'PREREQ':
                line_method = rr_line
                section = m.answer
            elif u == 'AUTHORITY' or u == 'UPDATE':
                line_method = rr_line
                section = m.authority
            elif u == 'ADDITIONAL':
                line_method = rr_line
                section = m.additional
            tok.get_eol()
            continue
        tok.unget(token)
        line_method(m, tok, section)
    return m

def from_text(text):
    """Convert the text format message into a message object."""
    return from_file(text)

def make_query(qname, rdtype, rdclass = DNS.rdataclass.IN):
    if isinstance(qname, str):
        qname = DNS.name.from_text(qname)
    if isinstance(rdtype, str):
        rdtype = DNS.rdatatype.from_text(rdtype)
    if isinstance(rdclass, str):
        rdclass = DNS.rdataclass.from_text(rdclass)
    m = Message()
    m.flags |= DNS.flags.RD
    n = m.find_node(m.question, qname, True)
    rds = n.find_rdataset(rdclass, rdtype, DNS.rdatatype.NONE, True, True)
    return m
